from __future__ import absolute_import
# Copyright 2014 Open Data Science Initiative and other authors. See AUTHORS.txt
# Licensed under the BSD 3-clause license (see LICENSE.txt)

import os
import sys

import pandas as pd
if sys.version_info>=(3,0):
    import urllib.parse as urlparse
else:
    import urlparse

from .config import *

import pickle
import pandas as pd
import pickle
import numpy as np

from . import config

gdata_available=True
try:
    import gdata.docs.client
    import gdata.spreadsheet.service # Requires python-gdata on Ubuntu
    import pods.google as gl
except ImportError:
    gdata_available=False


if gdata_available:
    default_class_dir = os.path.expanduser(os.path.expandvars(config.config.get('class info', 'dir')))

    class distributor():
        """
        Class for distributing google spreadsheets across the group for obtaining information.        
        :param spreadsheet_title: the title of the spreadsheet (used if the spreadsheet is created for the first time)
        :param worksheet_name: the name of the worksheet to create (used if the worksheet is created for the first time)
        :param column_indent: the column indent to use in the spreadsheet.
        :type column_indent: int
        :param gd_client: the google spreadsheet service client to use (default is None which performs a programmatic login)
        :param participant_list: the list of participants to who you want to distribute the documents.
        :type participant_list: either a string (for a filename) or a dictionary specifying a google doc and associated sheet number.
        """
        def __init__(self, spreadsheet_title='Google Spreadsheet', keys_file=None, participant_list=None, user_sep=',', class_dir=None, suffix=None, worksheet_name=None, gd_client=None, docs_client=None):

            self.gd_client = gd_client
            self.docs_client = docs_client

            self.worksheet_name = worksheet_name
            # suffix to apply to 'handles' if they are non-unique
            if suffix is None:
                self.suffix = '1'
            else:
                self.suffix = suffix

            # directory where information about the class is stored.
            if class_dir is None:
                class_dir = default_class_dir
            else:
                class_dir = os.path.expanduser(os.path.expandvars(class_dir))
            self.class_dir = class_dir
            self.spreadsheet_title = spreadsheet_title
            if keys_file is None:
                keys_file = 'spreadsheet_keys.pickle'
            if participant_list is None:
                participant_list = 'class_list.csv'
                
            if type(participant_list) is dict:
                # participant list is stored in a google doc
                if 'spreadsheet_key' in participant_list and 'worksheet_name' in participant_list:
                    self.participant_sheet = gl.sheet(spreadsheet_key=participant_list['spreadsheet_key'],
                                                      worksheet_name=participant_list['worksheet_name'], 
                                                      gd_client=self.gd_client, 
                                                      docs_client=self.docs_client)
                    # if gl.sheet had to login, store the details.
                    self.gd_client = self.participant_sheet.gd_client
                    self.docs_client = self.participant_sheet.docs_client
                    self.users = self.participant_sheet.read()
                    self.users.rename(columns={'Gmail Address': 'Email'}, inplace=True)
                else:
                    raise ValueError("If a the participant list is a dictionary, then it should encode a google doc for the particiant list with fields 'spreadsheet_key' and 'worksheet_name'.")

            elif type(participant_list) is str:
                # participant list is stored in a csv file.
                self.participant_list=os.path.join(self.class_dir, participant_list)
                self.users = pd.read_csv(self.participant_list, sep=user_sep)
            
            else:
                raise ValueError("Could not determine type of participant list.")

            # load spreadsheet keys if they exist.
            if keys_file is None:
                self.keys_file=os.path.join(self.class_dir, 'spreadsheet_keys.pickle')
            else:
                self.keys_file=os.path.join(self.class_dir, keys_file)
                
            if os.path.exists(self.keys_file):
                self.sheet_keys = pickle.load(open(self.keys_file, "rb"))
                # remove any spreadsheet keys that aren't in users list.
                remove_keys = []
                for key in self.sheet_keys:
                    if not np.any(self.users.Email==key):
                        remove_keys.append(key)
                for key in remove_keys:
                    del self.sheet_keys[key]
            else:
                self.sheet_keys = {}

        def _get_suffix(self):
            return self.suffix
        def _get_handle(self, user):
            """Get the name of a user with a given email."""
            ind = user==self.users.Email
            if ind.sum()==1:
                ind_val = self.users[ind].index[0]
                user_series = self.users.loc[ind_val]
                if 'Handle' in user_series:
                    return user_series.Handle
                elif 'handle' in user_series:
                    return user_series.handle
                elif 'Nickname' in user_series:
                    return user_series.Nickname
                elif 'nickname' in user_series:
                    return user_series.nickname
                elif 'name' in user_series:
                    return user_series.name
                elif 'Name' in user_series:
                    return user_series.Name
                elif 'Forename' in user_series:
                    if 'Surname' in user_series:
                        return user_series.Forename + ' ' + user_series.Surname
            else:
                return None

        def _delete_sheet(self, user):
            """Delete a user's sheet."""
            if self._sheet_exists(user):
                #sheet_key = self.sheet_keys[user]
                #sheet = gl.sheet(spreadsheet_key=self.sheet_keys[user])
                #sheet.share_delete()
                del self.sheet_keys[user]
                pickle.dump(self.sheet_keys, open(self.keys_file, "wb" ))

        def _sheet_exists(self, user):
            """Check if a sheet exists already."""
            if user in self.sheet_keys:
                return True
            else:
                return False
            

        def _get_sheet(self, user):
            """Get or create a sheet corresponding to a user."""
            # check if we've already got a spreadsheet for this user, otherwise create one.

            if user in self.sheet_keys:
                sheet = gl.sheet(spreadsheet_key=self.sheet_keys[user], 
                                 worksheet_name=self.worksheet_name,
                                 gd_client=self.gd_client, 
                                 docs_client=self.docs_client)
                # if gl.sheet had to login, store the details.
                self.gd_client = sheet.gd_client
                self.docs_client = sheet.docs_client
            else:
                name = self._get_handle(user)
                title = self.spreadsheet_title
                if name is not None:
                    title += ' ' + name
                sheet = gl.sheet(title=title, 
                                 gd_client=self.gd_client, 
                                 docs_client=self.docs_client)
                # if gl.sheet had to login, store the details.
                self.gd_client = sheet.gd_client
                self.docs_client = sheet.docs_client
                self.sheet_keys[user] = sheet._key
                pickle.dump(self.sheet_keys, open(self.keys_file, "wb" ))
            return sheet


        def write_body(self, data_frame, header=2, function=None):
            """Write body of data to the users' directories."""
            for user in self.users.Email:
                if function is not None:
                    udata = function(data_frame)
                else:
                    udata = data_frame
                sheet = self._get_sheet(user)
                # Write the data to the user's spreadsheet.
                sheet.write_body(udata, header=header)

        def write_comment(self, comment, row=1, column=1):
            """Write comments to the users' spreadsheets."""
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                # Write the film data to the user's spreadsheet.
                sheet.write_comment(comment, row, column)

        def write_headers(self, data_frame, header=None):
            """Write comments to the users' spreadsheets."""
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                # Write the film data to the user's spreadsheet.
                if header is None:
                    sheet.write_headers(data_frame)
                else:
                    sheet.write_headers(data_frame, header)
                
        def share(self, share_type='writer', send_notifications=False):
            """
            Share a document with a given list of users.
            """
            # share a document with the given list of users.
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                for share in sheet.share_list():
                    if user in share:
                        # TODO: change sharing here?
                        #sheet.share_modify(user, share_type, send_notifications=False)
                        return
                sheet.share([user], share_type, send_notifications)


        def share_delete(self):
            """
            Remove sharing from all the class.
            """
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                if user in sheet.share_list():
                    sheet.share_delete(user)

        def share_modify(self, share_type='reader', send_notifications=False):
            """
            :param user: email of the user to update.
            :type user: string
            :param share_type: type of sharing for the given user, type options are 'reader', 'writer', 'owner'
            :type user: string
            :param send_notifications: 
            """
            if share_type not in ['writer', 'reader', 'owner']:
                raise ValueError("Share type should be 'writer', 'reader' or 'owner'")
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                sheet.share_modify(user, share_type, send_notifications)


        def write(self, data_frame, header=None, comment=None, function=None, overwrite=False):
            """
            Write a pandas data frame to a google document. This function will overwrite existing cells, but will not clear them first.

            :param data_frame: the data frame to write.
            :type data_frame: pandas.DataFrame
            :param header: number of header rows in the document.
            :type header: int
            :param comment: a comment to make at the top of the document (requres header>1
            :type comment: str
            """
            for user in self.users.Email:
                if function is not None:
                    udata = function(data_frame)
                else:
                    udata = data_frame
                # Write the film data to the user's spreadsheet.
                if not self._sheet_exists(user) or overwrite:
                    sheet = self._get_sheet(user)
                    sheet.write(udata, header=header, comment=comment)

        def update(self, data_frame, columns=None, header=1, comment=None, overwrite=True):
            """
            Update a google document with a given data frame. The
            update function assumes that the columns of the data_frame and
            the google document match, and that an index in either the
            google document or the local data_frame identifies one row
            uniquely. If columns is provided as a list then only the
            listed columns are updated.

            **Notes**

            :data_frame : data frame to update the spreadsheet with.
            :type data_frame: pandas.DataFrame
            :param columns: which columns are updated in the spreadsheet (by default all columns are updated)
            :type columns: list
            :param hearder_rows: how many rows are in the header (including the column headers). By default there is 1 row. 
            :type header: int
            :param comment: comment to place in the top row of the header (requires header>1)
            :type comment: str
            :rtype: pandas.DataFrame

            .. Note:: Returns the data frame that was found in the spreadsheet.

            """
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                if function is not None:
                    udata = function(data_frame)
                else:
                    udata = data_frame
                sheet.update(udata, columns, header, comment, overwrite)


        def read(self, names=None, header=1, na_values=[], read_values=False, dtype={}, usecols=None):
            """
            Read in information from distributed Google documents storing entries. Fields present are defined in 'names'

            :param names: list of names to give to the columns (in case they aren't present in the spreadsheet). Default None (for None, the column headers are read from the spreadsheet.
            :type names: list
            :param header: number of rows to use as header (default is 1).
            :type header: int
            :param na_values: additional list containing entry types that are to be considered to be missing data (default is empty list).
            :type na_values: list
            :param read_values: whether to read values rather than the formulae in the spreadsheet (default is False).
            :type read_values: bool
            :param dtype: Type name or dict of column -> type Data type for data or columns. E.g. {'a': np.float64, 'b': np.int32}
            :type dtype: dictonary
            :param usecols: return a subset of the columns.
            :type usecols: list
            :return: a dictionary containing the results from each spreadsheet.
            :rtype: dict
            """
            data = {}
            for user in self.users.Email:
                sheet = self._get_sheet(user)
                handle = self._get_handle(user)
                while handle in data:
                    handle += self._get_suffix()
                data[handle] = sheet.read(names, header, na_values, read_values, dtype, usecols)
            return data

def download(name, course, github='SheffieldML/notebook/master/lab_classes/'):
    """Download a lab class from the relevant course
    :param course: the course short name to download the class from.
    :type course: string
    :param reference: reference to the course for downloading the class.
    :type reference: string
    :param github: github repo for downloading the course from.
    :type string: github repo for downloading the lab."""

    github_stub = 'https://raw.githubusercontent.com/'
    if not name.endswith('.ipynb'):
        name += '.ipynb'
    from pods.util import download_url
    download_url(os.path.join(github_stub, github, course, name), store_directory=course)



from __future__ import print_function
from __future__ import absolute_import
import sys
import os

from .config import *


def download_url(url, dir_name='.', save_name=None, store_directory=None, messages=True, suffix=''):
    """Download a file from a url and save it to disk."""
    if sys.version_info>=(3,0):
        from urllib.parse import quote
        from urllib.request import urlopen 
        from urllib.error import URLError
    else:
        from urllib2 import quote
        from urllib2 import urlopen
        from urllib2 import URLError
    i = url.rfind('/')
    file = url[i+1:]
    if store_directory is not None:
        dir_name = os.path.join(dir_name, store_directory)
    if save_name is None:
        save_name = file
    save_name = os.path.join(dir_name, save_name)
    print("Downloading ", url, "->", save_name)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    try:
        response = urlopen(url+suffix)
    except URLError(e):
        if not hasattr(e, "code"):
            raise
        if e.code > 399 and e.code<500:
            raise ValueError('Tried url ' + url + suffix + ' and received client error ' + str(e.code))
        elif e.code > 499:
            raise ValueError('Tried url ' + url + suffix + ' and received server error ' + str(e.code))
    with open(save_name, 'wb') as f:
        meta = response.info()
        content_length_str = meta.get("Content-Length")
        if content_length_str:
            if sys.version_info>=(3,0):
                file_size = int(content_length_str)
            else:
                file_size = int(content_length_str[0])
        else:
            file_size = None
        status = ""
        file_size_dl = 0
        block_sz = 8192
        line_length=30
        while True:
            buff = response.read(block_sz)
            if not buff:
                break
            file_size_dl += len(buff)
            f.write(buff)
            sys.stdout.write(" "*(len(status)) + "\r")
            if file_size:
                status = r"[{perc: <{ll}}] {dl:7.3f}/{full:.3f}MB".format(dl=file_size_dl/(1048576.),
                                                                       full=file_size/(1048576.), ll=line_length,
                                                                       perc="="*int(line_length*float(file_size_dl)/file_size))
            else:
                status = r"[{perc: <{ll}}] {dl:7.3f}MB".format(dl=file_size_dl/(1048576.),
                                                                       ll=line_length,
                                                                       perc="."*int(line_length*float(file_size_dl/(10*1048576.))))

            sys.stdout.write(status)
            sys.stdout.flush()
        sys.stdout.write(" "*(len(status)) + "\r")
        print(status)
        # if we wanted to get more sophisticated maybe we should check the response code here again even for successes.



# Copyright 2014 Open Data Science Initiative and other authors. See AUTHORS.txt
# Licensed under the BSD 3-clause license (see LICENSE.txt)
from __future__ import print_function
from __future__ import absolute_import

import sys
import os

# Deal with raw_input in python
try:
    input = raw_input
except NameError:
    pass
    
import pandas as pd

from .config import *
import numpy as np

from . import notebook as nb

gdata_available=True
try:
    import gdata.docs.client
    import gdata.spreadsheet.service 
except ImportError:
    gdata_available=False

if gdata_available:
    import atom

    class sheet():
        """
        Class for interchanging information between google spreadsheets and pandas data frames. The class manages a spreadsheet.

        :param spreadsheet_key: the google key of the spreadsheet to open (default is None which creates a new spreadsheet).
        :param worksheet_name: the worksheet in the spreadsheet to work with (default None which causes Sheet1 to be the name)
        :param title: the title of the spreadsheet (used if the spreadsheet is created for the first time)
        :param column_indent: the column indent to use in the spreadsheet.
        :type column_indent: int
        :param gd_client: the google spreadsheet service client to use (default is NOne which performs a programmatic login)
        :param docs_client: the google docs login client (default is none which causes a new client login)
        :param published: whether the google doc underlying the system has been published or not (default: False).
        :type published: bool
        """
        def __init__(self, spreadsheet_key=None, worksheet_name=None, title='Google Spreadsheet', column_indent=0, gd_client=None, docs_client=None, published=False):

            self.email = config.get('google docs', 'user')
            self.password = config.get('google docs', 'password')

            source = 'ODS Gdata Bot'                

            
            self.published = published

            if gd_client is None:
                self.gd_client = gdata.spreadsheet.service.SpreadsheetsService()
                self.gd_client.email = self.email
                self.gd_client.password = self.password
                self.gd_client.source = source
                self.gd_client.ProgrammaticLogin()
            else:
                self.gd_client=gd_client

            if docs_client is None:
                self.docs_client = gdata.docs.client.DocsClient()
                self.docs_client.client_login(self.email, self.password, source)
            else:
                self.docs_client = docs_client

            if spreadsheet_key is None:
                # need to create the document
                document = gdata.docs.data.Resource(type='spreadsheet', title=title)
                self.document = self.docs_client.create_resource(document)
                self._key = self.document.get_id().split("%3A")[1]
                # try to ensure that we can set title properly. Not sure why resource is not initially returned.
                self.document = self._get_resource_feed()
                self.feed = self._get_worksheet_feed()
                if worksheet_name is not None:
                    self.worksheet_name='Sheet1'
                    self.change_sheet_name(worksheet_name)

            else:
                # document exists already
                self._key = spreadsheet_key
                self.document = self._get_resource_feed()
                
            self.set_sheet_focus(worksheet_name=worksheet_name)

            self.column_indent = column_indent
            self.url = 'https://docs.google.com/spreadsheets/d/' + self._key + '/'


#############################################################################
# Place methods here that are really associated with individual worksheets. #
#############################################################################

        def change_sheet_name(self, title):
            """Change the title of the current worksheet to title."""
            for entry in self.feed.entry:
                if self.worksheet_name==entry.title.text:
                    entry.title=atom.Title(text=title)
                    self.gd_client.UpdateWorksheet(entry)
                    self.worksheet_name = title
                    return
            raise ValueError("Can't find worksheet " + self.worksheet_name + " to change the name in Google spreadsheet " + self.url)
                

        def set_sheet_focus(self, worksheet_name):
            """Set the current worksheet to the given name. If the name doesn't exist then create the sheet using sheet.add_sheet()"""
            self.update_sheet_list()
            # if the worksheet is set to None default to first sheet, warn if it's name is not "Sheet1".
            if worksheet_name is None:
                self.worksheet_name = self.feed.entry[0].title.text
                if len(self.feed.entry)>1 and self.worksheet_name != 'Sheet1':
                    print("Warning, multiple worksheets in this spreadsheet and no title specified. Assuming you are requesting the sheet called '{sheetname}'. To surpress this warning, please specify the sheet name.".format(sheetname=self.worksheet_name))
            else:
                if worksheet_name not in self.id_dict:
                    # create new worksheet here.
                    self.add_sheet(worksheet_name=worksheet_name)
                    self.worksheet_name = worksheet_name
                else:
                    self.worksheet_name = worksheet_name
            # Get list of ids from the spreadsheet
            self.worksheet_id = self.id_dict[self.worksheet_name]

        def add_sheet(self, worksheet_name, rows=100, columns=10):
            """Add a worksheet. To add and set to the current sheet use set_sheet_focus()."""
            self.gd_client.AddWorksheet(title=worksheet_name, row_count=rows, col_count=columns, key=self._key)
            self.update_sheet_list()

        def write(self, data_frame, header=None, comment=None):
            """
            Write a pandas data frame to a google document. This function will overwrite existing cells, but will not clear them first.

            :param data_frame: the data frame to write.
            :type data_frame: pandas.DataFrame
            :param header: number of header rows in the document.
            :type header: int
            :param comment: a comment to make at the top of the document (requres header>1
            :type comment: str
            """
            if comment is not None:
                if header is None:
                    header=2
                elif header==1:
                    raise ValueError('Comment will be overwritten by column headers')
                self.write_comment(comment)
            else:
                if header is None:
                    header=1
            self.write_headers(data_frame, header)
            self.write_body(data_frame, header)

        def augment(self, data_frame, columns, header=1, comment=None):
            """
            Augment is a special wrapper function for update that calls it
            with overwrite set to False. Use this command if you only want
            to make changes when the cell in the spreadsheet is empty.
            """
            self.update(data_frame, columns, header, comment, overwrite=False)

        def update(self, data_frame, columns=None, header=1, comment=None, overwrite=True):
            """
            Update a google document with a given data frame. The
            update function assumes that the columns of the data_frame and
            the google document match, and that an index in either the
            google document or the local data_frame identifies one row
            uniquely. If columns is provided as a list then only the
            listed columns are updated.

            **Notes**

            :data_frame : data frame to update the spreadsheet with.
            :type data_frame: pandas.DataFrame
            :param columns: which columns are updated in the spreadsheet (by default all columns are updated)
            :type columns: list
            :param hearder_rows: how many rows are in the header (including the column headers). By default there is 1 row. 
            :type header: int
            :param comment: comment to place in the top row of the header (requires header>1)
            :type comment: str
            :rtype: pandas.DataFrame

            .. Note:: Returns the data frame that was found in the spreadsheet.

            """
            if not data_frame.index.is_unique:
                raise ValueError("Index for data_frame is not unique in Google spreadsheet " + self.url)
            ss = self.read(header=header)
            if not ss.index.is_unique:
                raise ValueError("Index in google doc is not unique in Google spreadsheet " + self.url)
            if columns is None:
                columns = ss.columns
            if (len(set(ss.columns) - set(data_frame.columns))>0 or
                len(set(data_frame.columns) - set(ss.columns))>0):
                # TODO: Have a lazy option that doesn't mind this mismatch and accounts for it.
                raise ValueError('There is a mismatch between columns in online spreadsheet and the data frame we are using to update in Google spreadsheet ' + self.url)
            add_row = []
            remove_row = []
            update_cell = []
            # Compute necessary changes
            for index in data_frame.index:
                if index in ss.index:
                    for column in columns:
                        if overwrite:
                            if not ss[column][index] == data_frame[column][index]:
                                update_cell.append((index, column))
                        else:
                            if ((pd.isnull(ss[column][index]) 
                                or ss[column][index] == '') 
                                and not (pd.isnull(data_frame[column][index])
                                         or data_frame[column][index] == '')):
                                update_cell.append((index, column))

                else:
                    add_row.append(index)
            if overwrite:
                for index in ss.index:
                    if index not in data_frame.index:
                        remove_row.append(index)


            row_number = header+1

            self.row_batch_size = 10
            query = gdata.spreadsheet.service.CellQuery()
            query.return_empty = "true" 
            query.min_col = str(self.column_indent+1) 
            query.max_col = str(len(ss.columns)+1+self.column_indent)
            query.min_row = str(row_number)
            query.max_row = str(row_number + self.row_batch_size)
            cells = self._get_cell_feed(query=query)

            batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
            counter = 0
            row_to_delete = []
            row_number = 1
            for index, row in ss.iterrows():
                if counter>=len(cells.entry):
                    # Update current block
                    updated = self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)
                    # pull down new batch of cells.
                    query.min_row = str(row_number)
                    query.max_row = str(row_number + self.row_batch_size)
                    cells = self._get_cell_feed(query=query)
                    batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
                    counter = 0
                current_index = cells.entry[counter].cell.inputValue 
                row_number = int(cells.entry[counter].cell.row) + 1

                if current_index in remove_row:
                    v = int(cells.entry[counter].cell.row)-header-1
                    row_to_delete.append(v)
                    print("Warning deleting row indexed by '" + current_index + "' from " + cells.entry[counter].cell.row + " currently! Not comprehensively tested. Best guess is that row to delete is " + str(v) + " in Google spreadsheet " + self.url)
                    ans = input("Delete row (Y/N)?")
                    if len(ans)==0 or (not ans[0]=='Y' and not ans[0] == 'y'):
                        raise ValueError("Not willing to delete row.")
                    counter+=len(row)
                    continue
                else:
                    counter+=1
                for current_column, entry in row.items():
                    if (current_index, current_column) in update_cell:
                        val = data_frame[current_column][index]
                        if not pd.isnull(val):
                            v = []
                            try:
                                v = str(val)
                            except UnicodeDecodeError:
                                v = str(val)
                            cells.entry[counter].cell.inputValue = v
                        else:
                            cells.entry[counter].cell.inputValue = ''
                        batchRequest.AddUpdate(cells.entry[counter])
                    counter+=1
            updated = self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)
            # Delete the rows to be removed.
            for row in sorted(row_to_delete, reverse=True):
                print(("Delete row ", row))
                self._delete_row(row)
            # Insert the rows to be added
            for index in add_row:
                self._add_row(index, data_frame.loc[index])

        def _delete_row(self, row_number):
            """
            Delete a row of the spreadsheet.
            :param row_number: the row number to be deleted.
            :type row_number: int"""
            list_feed = self._get_list_feed(self._key, self.worksheet_id)
            self.gd_client.DeleteRow(list_feed.entry[row_number])

        def _add_row(self, index, data_series):
            """
            Add a row to the spreadsheet.
            :param index: index of the row to be added.
            :type index: str or int (any valid index for a pandas.DataFrame)
            :param data_series: the entries of the row to be added.
            :type data_series: pandas.Series"""
            dict = {}
            dict['index'] = index
            for column, entry in data_series.items():
                if not pd.isnull(entry):
                    val = []
                    try:
                        val = str(entry)
                    except UnicodeDecodeError:
                        val = str(entry)
                    dict[column] = val
            self.gd_client.InsertRow(dict, self._key, self.worksheet_id)


        def write_comment(self, comment, row=1, column=1):
            """Write a comment in the given cell"""
            query = gdata.spreadsheet.service.CellQuery()
            query.return_empty = "true" 
            query.min_col = str(column) 
            query.max_col = str(column)
            query.min_row = str(row)
            query.max_row = str(row)
            cells = self._get_cell_feed(query=query)
            batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
            cells.entry[0].cell.inputValue = comment
            batchRequest.AddUpdate(cells.entry[0])
            updated = self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)

        def write_body(self, data_frame, header):
            """Write the body of a data frame to a google doc."""
            # query needs to be set large enough to pull down relevant cells of sheet.
            row_number = header+1
            self.row_batch_size = 10

            query = gdata.spreadsheet.service.CellQuery()
            query.return_empty = "true" 
            query.min_col = str(self.column_indent+1) 
            query.max_col = str(len(data_frame.columns)+1+self.column_indent)
            query.min_row = str(row_number)
            query.max_row = str(row_number + self.row_batch_size)
            cells = self._get_cell_feed(query=query)
            batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
            counter = 0
            for index, row in data_frame.iterrows():
                if counter>=len(cells.entry):
                    # Update current block
                    row_number = int(cells.entry[counter-1].cell.row) + 1
                    updated = self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)
                    # pull down new batch of cells.
                    query.min_row = str(row_number)
                    query.max_row = str(row_number + self.row_batch_size)
                    cells = self._get_cell_feed(query=query)
                    batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
                    counter = 0
                cells.entry[counter].cell.inputValue = str(index)
                batchRequest.AddUpdate(cells.entry[counter])
                counter+=1
                for entry in row:
                    if not pd.isnull(entry):
                        val = []
                        try:
                            val = str(entry)
                        except UnicodeDecodeError:
                            val = str(entry)
                        cells.entry[counter].cell.inputValue = val
                    batchRequest.AddUpdate(cells.entry[counter])
                    counter+=1
            updated = self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)

        def write_headers(self, data_frame, header=1):
            """Write the headers of a data frame to the spreadsheet."""
            # query needs to be set large enough to pull down relevant cells of sheet.
            query = gdata.spreadsheet.service.CellQuery()
            query.return_empty = "true" 
            query.min_row = str(header)
            query.max_row = query.min_row
            query.min_col = str(self.column_indent+1) 
            query.max_col = str(len(data_frame.columns)+self.column_indent+1) 

            cells = self._get_cell_feed(query=query)

            batchRequest = gdata.spreadsheet.SpreadsheetsCellsFeed()
            index_name = data_frame.index.name 
            if index_name == '' or index_name is None:
                index_name = 'index'
            headers = [index_name] + list(data_frame.columns)
            for i, column in enumerate(headers):
                cells.entry[i].cell.inputValue = column
                batchRequest.AddUpdate(cells.entry[i])

            return self.gd_client.ExecuteBatch(batchRequest, cells.GetBatchLink().href)

        def read(self, names=None, header=1, na_values=[], read_values=False, dtype={}, usecols=None, index_field=None):
            """
            Read in information from a Google document storing entries. Fields present are defined in 'names'

            :param names: list of names to give to the columns (in case they aren't present in the spreadsheet). Default None (for None, the column headers are read from the spreadsheet.
            :type names: list
            :param header: number of rows to use as header (default is 1).
            :type header: int
            :param na_values: additional list containing entry types that are to be considered to be missing data (default is empty list).
            :type na_values: list
            :param read_values: whether to read values rather than the formulae in the spreadsheet (default is False).
            :type read_values: bool
            :param dtype: Type name or dict of column -> type Data type for data or columns. E.g. {'a': np.float64, 'b': np.int32}
            :type dtype: dictonary
            :param usecols: return a subset of the columns.
            :type usecols: list
            """

            # todo: need to check if something is written below the 'table' as this will be read (for example a rogue entry in the row below the last row of the data.
            entries_dict = {}
            index_dict = {}
            cells =  self._get_cell_feed()
            read_names = False
            if names is None:
                read_names = True
                names = {}
            for myentry in cells.entry:
                col = myentry.cell.col
                row = myentry.cell.row
                if read_values:
                    # Compute the evaluated cell entry
                    value = myentry.cell.value
                else:
                    # return a formula if it's present
                    value = myentry.cell.inputValue
                if int(row)<header:
                    continue
                if int(row)==header:
                    if read_names:
                        # Read the column titles for the fields.
                        fieldname = value.strip()
                        if fieldname in list(names.values()):
                            print(("ValueError, Field name duplicated in header in spreadsheet name:", self.worksheet_name, "in Google sheet ", self.url))
                            ans = input('Try and fix the error on the sheet and then return here. Error fixed (Y/N)?')
                            if ans[0]=='Y' or ans[0] == 'y':
                                return self.read(names, header, na_values, read_values, dtype, usecols, index_field)
                            raise ValueError("Field name duplicated in header in sheet in " + self.worksheet_name + " in Google spreadsheet " + self.url)
                        else:
                            names[col] = fieldname
                    continue

                if not row in entries_dict:
                    entries_dict[row] = {}

                if col in names:
                    field = names[col]
                else:
                    field = col

                if ((index_field is None and field.lower() == 'index') 
                    or field==index_field):
                    val = value.strip()
                    if field in dtype:
                        index_dict[row] = dtype[field](val)
                    else:
                        index_dict[row] = val
                elif usecols is None or field in usecols:
                    val = value.strip()
                    if not val in na_values and val != '':
                        if field in dtype:
                            entries_dict[row][field] = dtype[field](val)
                        else:
                            try:
                                a = float(val)
                            except ValueError:
                                entries_dict[row][field] = val
                            else:
                                try:
                                    a = int(a)
                                except ValueError:
                                    entries_dict[row][field] = a
                                else:
                                    if a == int(a):
                                        entries_dict[row][field] = int(a)
                                    else:
                                        entries_dict[row][field] = a



            entries = []
            index = []
            for key in sorted(list(entries_dict.keys()), key=int):
                entries.append(entries_dict[key])
                if len(index_dict)>0:
                    try:
                        index.append(index_dict[key])
                    except KeyError:
                        print(("KeyError, unidentified key in ", self.worksheet_name, " in Google spreadsheet ", self.url))
                        ans = input('Try and fix the error on the sheet and then return here. Error fixed (Y/N)?')
                        if ans[0]=='Y' or ans[0] == 'y':
                            return self.read(names, header, na_values, read_values, dtype, usecols, index_field)
                        else:
                            raise KeyError("Unidentified key in " + self.worksheet_name + " in Google spreadsheet " + self.url)

                else:
                    index.append(int(key)-header)

            if len(index)>0:
                entries = pd.DataFrame(entries, index=index)
            else:
                # this seems to cause problems, but it shouldn't come here now. If no index column is included index defaults to row number - header.
                entries = pd.DataFrame(entries)
            if len(names)>0:
                for field in list(names.values()):
                    if field not in list(entries.columns) + ['index']:
                        if usecols is None or field in usecols:
                            entries[field] = np.NaN

                column_order = []
                for key in sorted(names, key=int):
                    if usecols is None or names[key] in usecols:
                        column_order.append(names[key])
                if 'index' in column_order:
                    column_order.remove('index')
                entries = entries[column_order]

            return entries

#######################################################################
# Place methods here that are really associated with the spreadsheet. #
#######################################################################

        def set_title(self, title):
            """Change the title of the google spreadsheet."""
            pass
            #self.document.title = atom.data.Title(text=title)
            #self.docs_client.update_resource(self.document)

        def get_title(self):
            """Get the title of the google spreadsheet."""
            return self.document.title.text

        def delete_sheet(self, worksheet_name):
            """Delete the worksheet with the given name."""
            if worksheet_name == self.worksheet_name:
                raise ValueError("Can't delete the sheet I'm currently pointing to, use set_sheet_focus to change sheet first.")
            for entry in self.feed.entry:
                if worksheet_name==entry.title.text:
                    self.gd_client.DeleteWorksheet(entry)
                    return
            raise ValueError("Can't find worksheet " + worksheet_name + " to change the name " + " in Google spreadsheet " + self.url)

        def update_sheet_list(self):
            """Update object with the worksheet feed and the list of worksheet_ids, can only be run once there is a spreadsheet key and a resource feed in place. Needs to be rerun if a worksheet is added."""
            self.feed = self._get_worksheet_feed()
            self.id_dict = self.worksheet_ids()

        def _repr_html_(self):
            if self.published: #self.document.published.tag=='published':
                url = self.url + '/pubhtml?widget=true&amp;headers=false' 
                return nb.iframe_url(url, width=500, height=300)
            else:
                output = '<p><b>Google Sheet</b> at <a href="{url}" target="_blank">this url.</a>\n</p>'.format(url=self.url)
                return output + self.read()._repr_html_()
                #return None
            

        def show(self, width=400, height=200):
            """If the IPython notebook is available, and the google
            spreadsheet is published, then the spreadsheet is displayed
            centrally in a box."""
            if self.published:
                try:
                    from IPython.display import HTML
                    url = self.url + '/pubhtml?widget=true&amp;headers=false' 
                    nb.iframe_url(url, width=width, height=height)
                except ImportError:
                    print(ds.url)
                else:
                    raise
            
        def worksheet_ids(self):
            def _id(entry):
                split = urllib.parse.urlsplit(entry.id.text)
                return os.path.basename(split.path)
            return dict([(entry.title.text, _id(entry)) for entry in self.feed.entry])

        def share(self, users, share_type='writer', send_notifications=False):
            """
            Share a document with a given list of users.
            """
            # share a document with the given list of users.
            for user in users:
                acl_entry = gdata.docs.data.AclEntry(
                    scope=gdata.acl.data.AclScope(value=user, type='user'),
                    role=gdata.acl.data.AclRole(value=share_type),
                    )
                acl2 = self.docs_client.AddAclEntry(self.document, acl_entry, send_notifications=send_notifications)
                    

        def _get_acl_entry(self, user):
            """
            Return the acl entry associated with a given user name.
            """
            acl_feed = self._get_acl_feed()
            for acl_entry in acl_feed.entry:
                if acl_entry.scope.value == user:
                    return acl_entry
            raise ValueError("User: " + str(user) + " not in the acl feed for this resource" + " in Google spreadsheet " + self.url)


        def share_delete(self, user):
            """
            Remove sharing from a given user.
            """
            acl_entry = self._get_acl_entry(user)
            self.docs_client.DeleteAclEntry(acl_entry)

        def share_modify(self, user, share_type='reader', send_notifications=False):
            """
            :param user: email of the user to update.
            :type user: string
            :param share_type: type of sharing for the given user, type options are 'reader', 'writer', 'owner'
            :type user: string
            :param send_notifications: 
            """
            if share_type not in ['writer', 'reader', 'owner']:
                raise ValueError("Share type should be 'writer', 'reader' or 'owner'")

            #acl_entry = self._get_acl_entry(user)# update ACL entry

            #acl_entry.role.value = share_type
            # According to Ali Afshar you need to remove the etag (https://groups.google.com/forum/#!msg/google-documents-list-api/eFSmo14nDLA/Oo4SjePHZd8J), can't work out how though!
            #etag_element = acl_entry.find('etag')
            #acl_entry.remove(etagelement)
            #self.docs_client.UpdateAclEntry(acl_entry, sent_notifications=send_notifications)
            # Hack: delete and re-add.
            self.share_delete(user)
            self.share([user], share_type, send_notifications)

        def share_list(self):
            """
            Provide a list of all users who can access the document in the form of 
            """
            entries = []
            acl_feed = self._get_acl_feed()
            for acl_entry in acl_feed.entry:
                entries.append((acl_entry.scope.value, acl_entry.role.value))
            return entries

        def revision_history(self):
            """
            Get the revision history of the document from Google Docs.
            """
            for entry in self.docs_client.GetResources(limit=55).entry:
                revisions = self.docs_client.GetRevisions(entry)
                for revision in revisions.entry:
                    print((revision.publish, revision.GetPublishLink()))



        def _get_feed(self, type, query=None, tries=0, max_tries=10):
            """
            Check for exceptions when calling for a group of cells from
            the google docs API. Retry a maximum number of times (default 10).
            """
            try:
                if type == 'cell':
                    if query is None:
                        feed = self.gd_client.GetCellsFeed(self._key, wksht_id=self.worksheet_id)
                    else:
                        feed = self.gd_client.GetCellsFeed(self._key, wksht_id=self.worksheet_id, query=query)

                elif type == 'list':
                    feed = self.gd_client.GetListFeed(self._key, self.worksheet_id)
                elif type == 'worksheet':
                    feed = self.gd_client.GetWorksheetsFeed(self._key)

                elif type == 'acl':
                    feed = self.docs_client.GetAcl(self.document)

                elif type == 'resource':
                    feed = self.docs_client.GetResourceById(self._key)

            # Sometimes the server doesn't respond. Retry the request.
            except gdata.service.RequestError(inst):
                if tries<10:
                    status = inst[0]['status']
                    print("Error status: " + str(status) + '<br><br>' + inst[0]['reason'] + '<br><br>' + inst[0]['body'])
                    if status>499:
                        print(("Try", tries+1, "of", max_tries, "waiting 2 seconds and retrying."))
                        import sys
                        sys.stdout.flush()
                        import time
                        time.sleep(2)
                        feed = self._get_feed(type=type, query=query, tries=tries+1)
                    else:
                        raise 
                else:
                    print("Maximum tries at contacting Google servers exceeded.")
                    import sys
                    sys.stdout.flush()
                    raise
            return feed

        def _get_cell_feed(self, query=None, tries=0, max_tries=10):
            """
            Wrapper for _get_feed() when a cell feed is required.
            """
            # problem: if there are only 1000 lines in the spreadsheet and you request more you get this error: 400 Invalid query parameter value for max-row.
            return self._get_feed(type='cell', query=query, tries=tries, max_tries=max_tries)

        def _get_resource_feed(self, tries=0, max_tries=10):
            """
            Wrapper for _get_feed() when a resource is required.
            """
            return self._get_feed(type='resource', tries=tries, max_tries=max_tries)

        def _get_list_feed(self, tries=0, max_tries=10):
            """
            Wrapper for _get_feed() when a list feed is required.
            """
            return self._get_feed(type='list', query=None, tries=tries, max_tries=max_tries)

        def _get_worksheet_feed(self, tries=0, max_tries=10):
            """
            Wrapper for _get_feed() when a cell worksheet feed is required.
            """
            return self._get_feed(type='worksheet', query=None, tries=tries, max_tries=max_tries)

        def _get_acl_feed(self, tries=0, max_tries=10):
            """
            Wrapper for _get_feed() when an acl feed is required.
            """
            return self._get_feed(type='acl', query=None, tries=tries, max_tries=max_tries)



# Copyright 2014 Open Data Science Initiative and other authors. See AUTHORS.txt
# Licensed under the BSD 3-clause license (see LICENSE.txt)


def display_url(target):
    """Displaying URL in an IPython notebook to allow the user to click and check on information. With thanks to Fernando Perez for putting together the implementation!
    :param target: the url to display.
    :type target: string."""
    from IPython.display import display, HTML
    prefix = u"http://" if not target.startswith("http") else u""
    target = prefix + target
    display(HTML(u'<a href="{t}" target=_blank>{t}</a>'.format(t=target)))

def iframe_url(target, width=500, height=400, scrolling=True, border=0, frameborder=0):
    """Produce an iframe for displaying an item in HTML window.
    :param target: the target url.
    :type target: string
    :param width: the width of the iframe (default 500).
    :type width: int
    :param height: the height of the iframe (default 400).
    :type height: int
    :param scrolling: whether or not to allow scrolling (default True).
    :type scrolling: bool
    :param border: width of the border.
    :type border: int
    :param frameborder: width of the frameborder.
    :type frameborder: int"""
    
    prefix = u"http://" if not target.startswith("http") else u""
    target = prefix + target
    if scrolling:
        scroll_val = 'yes'
    else:
        scroll_val = 'no'
    return u'<iframe frameborder="{frameborder}" scrolling="{scrolling}" style="border:{border}px" src="{url}", width={width} height={height}></iframe>'.format(frameborder=frameborder, scrolling=scroll_val, border=border, url=target, width=width, height=height)

def display_iframe_url(target, **kwargs):
    """Display the contents of a URL in an IPython notebook.
    
    :param target: the target url.
    :type target: string

    .. seealso:: `iframe_url()` for additional arguments."""

    from IPython.display import display, HTML
    txt = iframe_url(target, **kwargs)
    display(HTML(txt))


def display_google_book(id, page, width=700, height=500, **kwargs):
    """Display an embedded version of a Google book.
    :param id: the id of the google book to display.
    :type id: string
    :param page: the start page for the book.
    :type id: string or int."""
    url = 'http://books.google.co.uk/books?id={id}&pg=PA{page}&output=embed'.format(id=id, page=page)
    display_iframe_url(url, width=width, height=height, **kwargs)

def code_hide():
    """Hide code in a notebook."""
def code_toggle(start_show=False, message=None):
    """Toggling on and off code in a notebook. 
    :param start_show: Whether to display the code or not on first load (default is False).
    :type start_show: bool
    :param message: the message used to toggle display of the code.
    :type message: string

    The tip that this idea is
    based on is from Damian Kao (http://blog.nextgenetics.net/?e=102)."""

    from IPython.display import display, HTML
    
    html ='<script>\n'
    if message is None:
        message = u'The raw code for this jupyter notebook can be hidden for easier reading.'
    if start_show:
        html += u'code_show=true;\n'
    else:
        html += u'code_show=false;\n'
    html+='''function code_toggle() {
 if (code_show){
 $('div.input').show();
 } else {
 $('div.input').hide();
 }
 code_show = !code_show
} 
$( document ).ready(code_toggle);
</script>
'''
    html += message + ' To toggle on/off the raw code, click <a href="javascript:code_toggle()">here</a>.' 
    display(HTML(html))


    
def display_prediction(basis, num_basis=4, wlim=(-1.,1.), fig=None, ax=None, xlim=None, ylim=None, num_points=1000, offset=0.0, **kwargs):
    """Interactive widget for displaying a prediction function based on summing separate basis functions.
    :param basis: a function handle that calls the basis functions.
    :type basis: function handle.
    :param xlim: limits of the x axis to use.
    :param ylim: limits of the y axis to use.
    :param wlim: limits for the basis function weights."""

    import numpy as np
    from IPython.html.widgets.interaction import interact, fixed
    from IPython.display import display
    import pylab as plt

    if fig is not None:
        if ax is None:
            ax = fig.gca()

    if xlim is None:
        if ax is not None:
            xlim = ax.get_xlim()
        else:
            xlim = (-2., 2.)
    if ylim is None:
        if ax is not None:
            ylim = ax.get_ylim()
        else:
            ylim = (-1., 1.)

    # initialise X and set up W arguments.
    x = np.zeros((num_points, 1))
    x[:, 0] = np.linspace(xlim[0], xlim[1], num_points)
    param_args = {}
    for i in range(num_basis):
        lim = list(wlim)
        if i ==0:
            lim[0] += offset
            lim[1] += offset
        param_args['w_' + str(i)] = lim

    # helper function for making basis prediction.
    def predict_basis(w, basis, x, num_basis, **kwargs):
        Phi = basis(x, num_basis, **kwargs)
        f = np.dot(Phi, w)
        return f, Phi
    
    if type(basis) is dict:
        use_basis = basis[list(basis.keys())[0]]
    else:
        use_basis = basis
    f, Phi = predict_basis(np.zeros((num_basis, 1)),
                           use_basis, x, num_basis,
                           **kwargs)
    if fig is None:
        fig, ax=plt.subplots(figsize=(12,4))
        ax.set_ylim(ylim)
        ax.set_xlim(xlim)

    predline = ax.plot(x, f, linewidth=2)[0]
    basislines = []
    for i in range(num_basis):
        basislines.append(ax.plot(x, Phi[:, i], 'r')[0])

    ax.set_ylim(ylim)
    ax.set_xlim(xlim)

    def generate_function(basis, num_basis, predline, basislines, basis_args, display_basis, offset, **kwargs):
        w = np.zeros((num_basis, 1))
        for i in range(num_basis):
            w[i] = kwargs['w_'+ str(i)]
        f, Phi = predict_basis(w, basis, x, num_basis, **basis_args)
        predline.set_xdata(x[:, 0])
        predline.set_ydata(f)
        for i in range(num_basis):
            basislines[i].set_xdata(x[:, 0])
            basislines[i].set_ydata(Phi[:, i])

        if display_basis:
            for i in range(num_basis):
                basislines[i].set_alpha(1) # make visible
        else:
            for i in range(num_basis):
                basislines[i].set_alpha(0) 
        display(fig)
    if type(basis) is not dict:
        basis = fixed(basis)

    plt.close(fig)
    interact(generate_function, 
             basis=basis,
             num_basis=fixed(num_basis),
             predline=fixed(predline),
             basislines=fixed(basislines),
             basis_args=fixed(kwargs),
             offset = fixed(offset),
             display_basis = False,
             **param_args)



#
# This loads the configuration
#
import sys
import os
    
if sys.version_info>=(3,0):
    import configparser
else:
    import ConfigParser as configparser

config = configparser.ConfigParser()

# This is the default configuration file that always needs to be present.
default_file = os.path.abspath(os.path.join(os.path.dirname( __file__ ), 'defaults.cfg'))

# These files are optional 
# This specifies configurations that are typically specific to the machine (it is found alongside the GPy installation).
local_file = os.path.abspath(os.path.join(os.path.dirname( __file__ ), 'machine.cfg'))

# This specifies configurations specific to the user (it is found in the user home directory)
home = os.getenv('HOME') or os.getenv('USERPROFILE')
user_file = os.path.join(home,'.ods_user.cfg')

# Read in the given files.
config.readfp(open(default_file))
config.read([local_file, user_file])
if not config:
    raise ValueError("No configuration file found at either " + user_file + " or " + local_file + " or " + default_file + ".")



class density_network(Model):
    """
    Density network.

    :param X: input observations
    :param Y: output observations
    :param kernel: a GPy kernel, defaults to rbf+white
    :param likelihood: a GPy likelihood
    :param inference_method: The :class:`~GPy.inference.latent_function_inference.LatentFunctionInference` inference method to use for this GP
    :rtype: model object
    :param Norm normalizer:
        normalize the outputs Y.
        Prediction will be un-normalized using this normalizer.
        If normalizer is None, we will normalize using MeanNorm.
        If normalizer is False, no normalization will be done.

    .. Note:: Multiple independent outputs are allowed using columns of Y


    """
    def __init__(self, input_dim, Y, options, name='dnet'):
        super(density_network, self).__init__(name)
        assert Y.ndim == 2
        self.Y = Y.copy()
        self.num_data, self.output_dim = self.Y.shape

        logger.info("initializing X")
    
    def log_likelihood(self):

function ll = dnetLogLikelihood(model)

% DNETLOGLIKELIHOOD Density network log likelihood.
% FORMAT
% DESC computes the log likelihood of a density network
% model. 
% ARG model : the model structure for computing the log likelihood.
% RETURN ll : the model log likelihood.
%
% SEEALSO : modelLogLikeihood
%
% COPYRIGHT : Neil D. Lawrence, 2008

% MLTOOLS

  
ll = 0.5*model.d*model.N*log(model.beta/(2*pi)) ...
     - model.N*log(model.M); 

ll = ll - 0.5*model.alpha*sum(sum(model.A.*model.A)) ...
     - 0.5*model.alpha*sum(model.b.*model.b);

  
llPointComp = zeros(model.N, model.M);
% Get projections of latent samples.
Ypred = dnetOut(model);

if model.N > model.M
  for i = 1:model.M
    diffY = model.y - repmat(Ypred(i, :), model.N, 1);
    diffY = diffY.*diffY;
    llPointComp(:, i) = - 0.5*model.beta*sum(diffY, 2);
  end
else
  for i = 1:model.N
    diffY = repmat(model.y(i, :), model.M, 1) - Ypred;
    diffY = diffY.*diffY;
    llPointComp(i, :) = - 0.5*model.beta*sum(diffY, 2)';
  end
end

maxllPointComp = max(llPointComp, [], 2);
llPointComp = exp(llPointComp - repmat(maxllPointComp, 1, model.M));
ll = ll + sum(log(sum(llPointComp, 2)) + maxllPointComp);



from __future__ import print_function
from __future__ import absolute_import
import smtplib
import getpass

from . import config

gmail_sender= config.config.get('Gmail', 'user')
gmail_name = config.config.get('Gmail', 'name')

class gmail():
    def __init__(self, username=None, name=None):
        """Open an SMTP TLS session for gmail."""
        self.session = smtplib.SMTP('smtp.gmail.com', 587)
        self.session.ehlo()
        self.session.starttls()
        if name is None:
            self.name = gmail_name
        else:
            self.name = name
        if username is None:
            self.username = gmail_sender
        else:
            self.username = username
        self.password = None
        self.get_password()
        self.session.login(self.username,self.password)

    def get_password(self):
        if self.password is None:
            print("Check console for password input!")
            import sys
            sys.stdout.flush()
            self.password = getpass.getpass("Enter your password for gmail:")
        

    def send(self, recipient, subject, body, cc=[], reply_to=None, html=True):
        # The below code never changes, though obviously those
        # variables need values.
        if self.name is not None:
            sender = self.name + '<' + self.username + '>'
        else:
            sender = self.username
        if reply_to is None:
            reply_to = sender
        email_cc = ';'.join(cc)
        if html:
            content_type = 'text/html'
        else:
            content_type = 'text/plain'
        headers = "\r\n".join(["from: " + sender,
                               "subject: " + subject,
                               "cc: " + email_cc,
                               "to: " + recipient,
                                "reply-to: " + reply_to,
                               "mime-version: 1.0",
                               "content-type: " + content_type])

        # body_of_email can be plaintext or html!                    
        content = headers + "\r\n\r\n" + body
        self.session.sendmail(self.username, recipient, content)



from __future__ import absolute_import
import pandas as pd
import numpy as np
import json
from .config import config
import os

check_mark = '<span style="color:red;">**&#10004;**</span>'
check_mark = '<span style="color:red;">**Correct**</span>'

short_name = config.get('assesser', 'assessment_short_name')
long_name = config.get('assesser', 'assessment_long_name')
year = config.get('assesser', 'assessment_year')

data_directory = os.path.expandvars(config.get('assesser', 'data_directory'))

instructor_email = config.get('assesser', 'instructor_email')
instructor_name = config.get('assesser', 'instructor_name')
assesser_group_email = config.get('assesser', 'assesser_group_email')

participant_key = config.get('assesser', 'participant_list_key')
participant_sheet = config.get('assesser', 'participant_list_sheet')
marksheets_filename = config.get('assesser', 'class_marksheets_pickle')

class assessment():
    """Class for storing assesment information. This class stores the
    questions and answers for an assessment paper. It
    produces mark sheets and provides individual feedback.

    :param source: source of the exam information. If a dict then keys
    for 'spreadsheet_key' and 'worksheet_name' are expected. Questions
    will be sought in the associated google sheet. If it's a string
    that ends with 'json' then a json file will be loaded. If it's a
    string that ends with '.pickle' then a pickle file will be loaded,
    if it's a string that ends with '.csv' then a csv will be loaded.
    :type source: dict or str

    """
    
    def __init__(self, part=0, source=None, answers_sep=None, display_answer=False, display_marks=True):
        self.display_answer=display_answer
        self.display_marks=display_marks

        self.latex_preamble = r"""\documentclass[]{article}
\usepackage{a4}
\usepackage{amssymb,amsmath}
\usepackage{ifxetex,ifluatex}
\usepackage[utf8]{inputenc}
\usepackage{eurosym}
\usepackage{color}
\usepackage{fancyvrb}
\DefineShortVerb[commandchars=\\\{\}]{\|}
\DefineVerbatimEnvironment{Highlighting}{Verbatim}{commandchars=\\\{\}}
% Add ',fontsize=\small' for more characters per line
\newenvironment{Shaded}{}{}
\newcommand{\KeywordTok}[1]{\textcolor[rgb]{0.00,0.44,0.13}{\textbf{{#1}}}}
\newcommand{\DataTypeTok}[1]{\textcolor[rgb]{0.56,0.13,0.00}{{#1}}}
\newcommand{\DecValTok}[1]{\textcolor[rgb]{0.25,0.63,0.44}{{#1}}}
\newcommand{\BaseNTok}[1]{\textcolor[rgb]{0.25,0.63,0.44}{{#1}}}
\newcommand{\FloatTok}[1]{\textcolor[rgb]{0.25,0.63,0.44}{{#1}}}
\newcommand{\CharTok}[1]{\textcolor[rgb]{0.25,0.44,0.63}{{#1}}}
\newcommand{\StringTok}[1]{\textcolor[rgb]{0.25,0.44,0.63}{{#1}}}
\newcommand{\CommentTok}[1]{\textcolor[rgb]{0.38,0.63,0.69}{\textit{{#1}}}}
\newcommand{\OtherTok}[1]{\textcolor[rgb]{0.00,0.44,0.13}{{#1}}}
\newcommand{\AlertTok}[1]{\textcolor[rgb]{1.00,0.00,0.00}{\textbf{{#1}}}}
\newcommand{\FunctionTok}[1]{\textcolor[rgb]{0.02,0.16,0.49}{{#1}}}
\newcommand{\RegionMarkerTok}[1]{{#1}}
\newcommand{\ErrorTok}[1]{\textcolor[rgb]{1.00,0.00,0.00}{\textbf{{#1}}}}
\newcommand{\NormalTok}[1]{{#1}}
\ifxetex
  \usepackage[setpagesize=false, % page size defined by xetex
              unicode=false, % unicode breaks when used with xetex
              xetex,
              colorlinks=true,
              linkcolor=blue]{hyperref}
\else
  \usepackage[unicode=true,
              colorlinks=true,
              linkcolor=blue]{hyperref}
\fi
\hypersetup{breaklinks=true, pdfborder={0 0 0}}
\setlength{\parindent}{0pt}
\setlength{\parskip}{6pt plus 2pt minus 1pt}
\setlength{\emergencystretch}{3em}  % prevent overfull lines
\setcounter{secnumdepth}{0}
 

\begin{document}
\title{Lab Assessment """ + str(part+1) + "}\n"
        self.latex_post = "\n" + r"""\end{document}"""


        self.html_preamble = r"""<html>
<head>
  <title>Lab Assessment {part}</title>
  <script type="text/javascript"
    src="http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_HTMLorMML">
  </script>
</head>
<body>
  <h1>Lab Assessment {part}</h1>
""".format(part=part+1)
        self.html_post = r"""  </body>
</html>"""
        # if source is None:
        #     raise ValueError("You need to provide a source for the assesment question and answers.")
        # elif type(source) is dict:
        #     # participant list is stored in a google doc
        #     if 'spreadsheet_key' in source and 'worksheet_name' in source:
        #         self.participant_sheet = gl.sheet(spreadsheet_key=source['spreadsheet_key'],
        #                                           worksheet_name=source['worksheet_name'], 
        #                                           gd_client=self.gd_client, 
        #                                           docs_client=self.docs_client)
        #         # if gl.sheet had to login, store the details.
        #         self.gd_client = self.participant_sheet.gd_client
        #         self.docs_client = self.participant_sheet.docs_client
        #         self.answers = self.participant_sheet.read()
        #         self.users.rename(columns={'Gmail Address': 'Email'}, 
        #                           inplace=True)
        #     else:
        #         raise ValueError, "If a the source is a dictionary, then it should encode a google doc for the source with fields 'spreadsheet_key' and 'worksheet_name'."

        # elif type(source) is str:
        #     # source list is stored in a file.
        #     if target.endswith(".csv"):
        #         self.source=os.path.join(self.class_dir, source)
        #         self.answers = pd.read_csv(self.source, sep=answers_sep)
        #     elif target.endswidth(".pickle"):
        #         import pickle
        #         self.source=os.path.join(self.class_dir, source)
        #         self.answers = pickle.load(open(self.source, "rb"))
        #     elif target.endswidth(".json"):
        #         import json
        #         self.source=os.path.join(self.class_dir, source)
        #         self.answers = json.load(open(self.source, "rb"))
                
        # else:
        #     raise ValueError, "Could not determine type of source file."
        self.answers = answer(part)
            
    def latex(self):
        """Gives a latex representation of the assessment."""
        output = self.latex_preamble
        output += self._repr_latex_()
        output += self.latex_post
        return output

    def html(self):
        """Gives an html representation of the assessment."""
        output = self.html_preamble
        output += self._repr_html_()
        output += self.html_post
        return output


    def _repr_html_(self):
        from IPython.nbconvert.filters.markdown import markdown2html
        return markdown2html(self._repr_md_())

    def _repr_latex_(self):
        from IPython.nbconvert.filters.markdown import markdown2latex
        return markdown2latex(self._repr_md_())

    def _repr_md_(self):
        output = ''
        
        for qu, answer in enumerate(self.answers):
            output += '#### Assignment Question ' + str(qu+1) + '\n\n'
            mark = 0
            question = ''
            for number, part in enumerate(answer):
                if number==0:
                    # First part is the intro ramble
                    if type(part) is list or type(part) is tuple:
                        output += '' + part[0] + '\n\n'
                        if len(part)>1 and part[1] is False:
                            display_subs = False
                        else:
                            display_subs = True
                    elif len(part)>0:
                        output += '' + part + '\n\n'
                        display_subs = True
                    else:
                        display_subs = True

                else:                    
                    if display_subs:
                        question += ' ' + part[0] 
                    mark += part[2]
                    if (len(part)>3 and part[3] and not self.display_answer):
                        pass
                    else:
                        output += question
                        if self.display_answer:
                            if part[1] == check_mark:
                                output += ' '
                            else:
                                output += '\n\n'
                        question = ''
                        if self.display_answer:
                            output += '\n\n**Answer**\n\n' + part[1] + '\n\n'
                        if part[2]>0 and self.display_marks:
                            output += ' *' + str(mark) + ' marks*\n\n'
                        mark=0
        return output
    
    def marksheet(self):
        """Returns an pandas empty dataframe object containing rows and columns for marking. This can then be passed to a google doc that is distributed to markers for editing with the mark for each section."""
        columns=['Number', 'Question', 'Correct (a fraction)', 'Max Mark', 'Comments']
        mark_sheet = pd.DataFrame() 
        for qu_number, question in enumerate(self.answers):
            part_no = 0
            for number, part in enumerate(question):
                if number>0:
                    if part[2] > 0:
                        part_no += 1
                        index = str(qu_number+1) +'_'+str(part_no)
                        frame = pd.DataFrame(columns=columns, index=[index])
                        frame.loc[index]['Number'] = index
                        frame.loc[index]['Question'] = part[0]
                        frame.loc[index]['Max Mark'] = part[2]
                        mark_sheet =  mark_sheet.append(frame)

        return mark_sheet.sort(columns='Number')

    def total_marks(self):
        """Compute the total mark for the assessment."""
        total = 0
        for answer in self.answers:
            for number, part in enumerate(answer):
                if number>0:
                    if part[2]>0:
                        total+=part[2]
        return total

class feedback(assessment):
    """A class for providing student feedback."""
    def __init__(self, marksheet, part=0):
        assessment.__init__(self, part, display_answer=True, display_marks=False)
        self.marksheet=marksheet
        for qu in xrange(len(self.answers)):
            mark = 0
            question = ''
            counter = 0
            for number in xrange(len(self.answers[qu])):

                if number==0:
                    if type(self.answers[qu][number]) is list or type(self.answers[qu][number]) is tuple:
                        if len(self.answers[qu][number])>1 and self.answers[qu][number][1] is False:
                            self.answers[qu][number][1] = True
                else:  
                    mark = self.answers[qu][number][2]
                    if not np.isnan(mark) and mark > 0:
                        counter += 1
                        index = str(qu+1) + '_' + str(counter)
                        fraction = marksheet.loc[index]['Correct (a fraction)']
                        if fraction<1:
                            # Markers might write unicode in their
                            # comments, need to handle this here.
                            comment = marksheet.loc[index]['Comments']
                            
                            if type(comment) is float:
                                if np.isnan(comment):
                                    comment = ''
                            else:
                                if str(comment) != 'nan':
                                    comment = '<span style="color:red;">*Marker comment*:' + ' ' + str(comment) + '</span>\n\n'
                                else:
                                    comment = ''
                            if self.answers[qu][number][1] == '':
                                self.answers[qu][number][1] = '<span style="color:red;">No **-' + str((1-fraction)*mark) + ' mark(s)**</span> ' + comment
                            else:
                                self.answers[qu][number][1] = '' + comment + '<span style="color:red;">**Correct Answer:**</span>\n\n' + self.answers[qu][number][1] + '\n\n<span style="color:red;">**-' + str((1-fraction)*mark) + ' mark(s)**</span>'
                        else:
                            self.answers[qu][number][1] = check_mark
    def _repr_md_(self):
        import numpy as np
        total = np.ceil((self.marksheet['Correct (a fraction)']*self.marksheet['Max Mark']).sum())
        if np.isnan(total):
            return '### We do not have a record of a submitted assignment.\n\n#### Total Mark: 0'
        else:
            return '### <span style="color:red;">Total Marks ' + str(total) + '</span>\n\n' + assessment._repr_md_(self) 

        
def answer(part, module='mlai2014.json'):
    """Returns the answers to the lab classes."""
    marks = json.load(open(os.path.join(data_directory, module), 'rb'))
    return marks['Lab '  + str(part+1)]



from distutils.core import setup
setup(
  name = 'pods',
  packages = ['pods'], # this must be the same as the name above
  package_data={'pods': ['defaults.cfg', 'data_resources.json', 'football_teams.json']},
  version = 'v0.0.2-alpha',
  description = 'Python software for Open Data Science',
  author = 'The Open Data Science Initiative',
  author_email = 'lawrennd@gmail.com',
  url = 'https://github.com/sods/ods', # use the URL to the github repo
  download_url = 'https://github.com/sods/ods/tarball/v0.0.2-alpha', 
  keywords = ['datascience', 'open', 'google sheet'], 
  classifiers = [],
)
